"""
AST for Jac
"""
from antlr4 import InputStream, CommonTokenStream, ParseTreeListener
from antlr4.error.ErrorListener import ErrorListener

from jaseci.utils.utils import logger, parse_str_token
from jaseci.jac.jac_parse.jacLexer import jacLexer
from jaseci.jac.jac_parse.jacParser import jacParser, ParseTreeWalker
import os


class ast:
    """
    AST Nodes

    The kind field is used to represent the grammar rule
    TODO: Error handling if jac program has errors
    """

    _ast_head_map = {}

    def __init__(
        self,
        mod_name,
        mod_dir="./",
        jac_text=None,
        parse_errors=None,
        start_rule="start",
        fresh_start=True,
    ):
        if fresh_start:
            ast._ast_head_map = {}
        self.name = "unparsed"
        self.kind = "unparsed"
        self.context = {}
        self._parse_errors = parse_errors if parse_errors else []
        self._start_rule = start_rule
        self.mod_name = mod_name if mod_name is not None else "@default"
        self._mod_dir = mod_dir
        self._keep = False
        self.line = 0
        self.column = 0
        self.kid = []
        if jac_text:
            self.parse_jac_str(jac_text)

    def is_rule(self):
        """Returns true if node is a rule"""
        return self.kind == "rule"

    def is_terminal(self):
        """Returns true if node is a terminal"""
        return self.kind == "terminal"

    def token(self):
        if not self.is_terminal():
            logger.error(str(f"Non terminals (rules) don't have token info - {self}"))
            return None
        else:
            return self.context["token"]

    def token_text(self):
        if self.is_terminal():
            return self.token()["text"]

    def token_symbol(self):
        if self.is_terminal():
            return self.token()["symbol"]

    def parse_jac_str(self, jac_str):
        """Parse language and build ast from string"""
        ast._ast_head_map[self.mod_name] = self
        input_stream = InputStream(jac_str)
        lexer = jacLexer(input_stream)
        stream = CommonTokenStream(lexer)
        errors = self.jac_tree_error(self)
        parser = jacParser(stream)
        parser.removeErrorListeners()
        parser.addErrorListener(errors)
        tree = getattr(parser, self._start_rule)()
        builder = self.jac_tree_builder(self)
        walker = ParseTreeWalker()
        walker.walk(builder, tree)

        if self._parse_errors:
            logger.error(str(f"Parse errors encountered - {self}"))

    def __str__(self):
        res = f"{self.kind}:{self.name}"
        if self.is_terminal():
            res += f':{self.context["token"]["text"]}'
        return res

    def __repr__(self):
        return self.__str__()

    def get_tokens(self):
        """Return list of all tokens derived from this ast node"""
        tokens = []
        if "token" in self.context:
            tokens.append(self.context["token"])
            return tokens
        else:
            for i in self.kid:
                tokens = tokens + i.get_tokens()
        return tokens

    def get_text(self):
        """Get source text for this node"""
        ret = ""
        for i in self.get_tokens():
            ret += f"{i['text']} "
        return ret

    class jac_tree_builder(ParseTreeListener):
        """Converter class from Antlr trees to Jaseci Tree"""

        def __init__(self, tree_root):
            self.tree_root = tree_root
            self.node_stack = []

        def run_import_module(self, jac_ast):
            """
            import_module:
                KW_IMPORT LBRACE (import_items | '*') RBRACE
                KW_WITH STRING SEMI;

            TODO: Check for duplicate imports and ignore if imported
            """
            kid = jac_ast.kid
            fn = os.path.join(
                self.tree_root._mod_dir, parse_str_token(kid[-2].token_text())
            )
            mod_name = os.path.basename(fn)
            from_mod = self.tree_root.mod_name
            logger.debug(f"Importing items from {mod_name} to {from_mod}...")
            parsed_ast = None
            if mod_name in ast._ast_head_map.keys():
                parsed_ast = ast._ast_head_map[mod_name]
            elif os.path.isfile(fn):
                with open(fn, "r") as file:
                    jac_text = file.read()
                mdir = os.path.dirname(os.path.realpath(fn)) + "/"
                parsed_ast = ast(
                    jac_text=jac_text,
                    mod_name=mod_name,
                    mod_dir=mdir,
                    fresh_start=False,
                )
            else:
                err = (
                    f"Module not found for import! {mod_name} from"
                    + f" {from_mod} - {fn}"
                )
                self.tree_root._parse_errors.append(err)
            if parsed_ast:
                self.tree_root._parse_errors += parsed_ast._parse_errors
                import_elements = list(
                    filter(lambda x: x.name == "element", parsed_ast.kid)
                )
                if kid[2].name == "STAR_MUL":
                    ret = import_elements
                else:
                    ret = self.run_import_items(kid[2], import_elements)
                for i in ret:
                    i._keep = True
                return ret
            else:
                return []

        def run_import_items(self, jac_ast, import_elements):
            """
            import_items:
                KW_WALKER (STAR_MUL | import_names) (COMMA import_items)?
                | KW_NODE (STAR_MUL | import_names) (COMMA import_items)?
                | KW_EDGE (STAR_MUL | import_names) (COMMA import_items)?
                | KW_GRAPH (STAR_MUL | import_names) (COMMA import_items)?
                | KW_GLOBAL (STAR_MUL | import_names) (COMMA import_items)?;
            """
            kid = jac_ast.kid
            ret_elements = list(
                filter(
                    lambda x: x._keep or x.kid[0].kid[0].name == kid[0].name,
                    import_elements,
                )
            )
            if kid[1].name == "import_names":
                import_names = self.run_import_names(kid[1])
                ret_elements = list(
                    filter(
                        lambda x: x._keep
                        or x.kid[0].kid[1].token_text() in import_names,
                        ret_elements,
                    )
                )
                if len(ret_elements) < len(import_names):
                    err = (
                        f"{kid[1].mod_name}: Line {kid[1].line}: "
                        + "Module name not found!"
                    )
                    self.tree_root._parse_errors.append(err)

            if kid[-1].name == "import_items":
                return ret_elements + self.run_import_items(kid[-1], import_elements)

            return ret_elements

        def run_import_names(self, jac_ast):
            """
            import_names:
                DBL_COLON NAME
                | DBL_COLON LBRACE name_list RBRACE;
            """
            kid = jac_ast.kid
            if kid[1].name == "NAME":
                return [kid[1].token_text()]
            else:
                return self.run_name_list(kid[2])

        def run_name_list(self, jac_ast):
            """
            name_list: NAME (COMMA NAME)*;
            """
            kid = jac_ast.kid
            ret = []
            for i in kid:
                if i.name == "NAME":
                    ret.append(i.token_text())
            return ret

        def enterEveryRule(self, ctx):
            """Visits every node in antlr parse tree"""
            if len(self.node_stack) == 0:
                new_node = self.tree_root
            else:
                new_node = ast(
                    mod_name=self.tree_root.mod_name,
                    mod_dir=self.tree_root._mod_dir,
                    fresh_start=False,
                )
            new_node.name = jacParser.ruleNames[ctx.getRuleIndex()]
            new_node.kind = "rule"
            new_node._parse_errors = self.tree_root._parse_errors
            new_node.line = ctx.start.line
            new_node.column = ctx.start.column

            if len(self.node_stack) and new_node.name != "import_module":
                self.node_stack[-1].kid.append(new_node)
            self.node_stack.append(new_node)

        def exitEveryRule(self, ctx):
            """Overloaded function that visits every node on exit"""
            top = self.node_stack.pop()
            if top.name == "import_module":
                for i in self.run_import_module(top):
                    if i not in self.node_stack[-1].kid:
                        self.node_stack[-1].kid.append(i)

        def visitTerminal(self, node):
            """Visits terminals as walker walks, adds ast node"""
            new_node = ast(
                mod_name=self.tree_root.mod_name,
                mod_dir=self.tree_root._mod_dir,
                fresh_start=False,
            )
            new_node.name = jacParser.symbolicNames[node.getSymbol().type]
            new_node.kind = "terminal"
            new_node.line = node.getSymbol().line
            new_node.column = node.getSymbol().column

            token = {
                "symbol": jacParser.symbolicNames[node.getSymbol().type],
                "text": node.getSymbol().text,
            }
            new_node.context["token"] = token

            self.node_stack[-1].kid.append(new_node)

    class jac_tree_error(ErrorListener):
        """Accumulate errors as parse tree is walked"""

        def __init__(self, tree_root):
            self.tree_root = tree_root

        def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
            """Add error to error list"""
            self.tree_root._parse_errors.append(
                f"{str(self.tree_root.mod_name)}: line {str(line)}:"
                f"{str(column)} - {self.tree_root} - {msg}"
            )



"""
Sentinel interpreter for jac code in AST form

This interpreter should be inhereted from the class that manages state
referenced through self.
"""
from jaseci.actor.architype import architype
from jaseci.actor.walker import walker
from jaseci.jac.interpreter.interp import interp
from jaseci.utils.utils import parse_str_token
from jaseci.jac.ir.jac_code import jac_ast_to_ir
from jaseci.jac.machine.jac_scope import jac_scope


class sentinel_interp(interp):
    """Jac interpreter mixin for objects that will execute Jac code"""

    def run_start(self, jac_ast):
        """
        start: ver_label? element* EOF;
        """
        kid = self.set_cur_ast(jac_ast)
        if kid[0].name == "ver_label":
            self.run_ver_label(kid[0])
        for i in kid[:-1]:
            self.run_element(i)

    def run_ver_label(self, jac_ast):
        """
        ver_label: 'version' COLON STRING;
        """
        kid = self.set_cur_ast(jac_ast)
        self.version = parse_str_token(kid[2].token_text())

    def run_element(self, jac_ast):
        """
        element: global | architype | walker | test;
        """
        kid = self.set_cur_ast(jac_ast)
        if kid[0].name == "global_var":
            self.load_global_var(kid[0])
        elif kid[0].name == "architype":
            self.load_architype(kid[0])
        elif kid[0].name == "walker":
            self.load_walker(kid[0])
        elif kid[0].name == "test":
            self.load_test(kid[0])

    def load_global_var(self, jac_ast):
        """
        global_var:
            KW_GLOBAL NAME EQ expression (COMMA NAME EQ expression)* SEMI;
        """
        kid = self.set_cur_ast(jac_ast)
        kid = kid[1:]
        while kid[0].name == "NAME":
            const_name = kid[0].token_text()
            if const_name in self.global_vars:
                self.rt_error(f"Global {const_name} already defined!", kid[0])
            else:
                self.global_vars[const_name] = self.run_expression(kid[2]).value
            kid = kid[4:] if kid[3].name == "COMMA" else kid[3:]

    def load_architype(self, jac_ast):
        """
        architype:
            KW_NODE NAME (COLON NAME)* (COLON INT)? attr_block
            | KW_EDGE NAME (COLON NAME)* attr_block
            | KW_GRAPH NAME graph_block;
        """
        kid = self.set_cur_ast(jac_ast)
        name = kid[1].token_text()
        kind = kid[0].token_text()
        arch = architype(
            m_id=self._m_id, h=self._h, code_ir=jac_ast, name=name, kind=kind
        )
        if len(kid) > 2 and kid[2].name == "COLON":
            for i in kid[2:]:
                if i.name == "NAME":
                    arch.super_archs.append(i.token_text())
        if self.arch_ids.has_obj_by_name(arch.name, kind=arch.kind):
            self.arch_ids.destroy_obj_by_name(arch.name, kind=arch.kind)
        self.arch_ids.add_obj(arch)
        self.arch_can_compile(kid[-1], arch)
        return arch

    def arch_can_compile(self, jac_ast, arch):
        """Helper function to statically compile can stmts for arch"""
        kid = self.set_cur_ast(jac_ast)
        self.push_scope(jac_scope(parent=self, has_obj=self, action_sets=[]))
        if jac_ast.name == "attr_block":
            for i in kid:
                if i.name == "attr_stmt" and i.kid[0].name == "can_stmt":
                    self.run_can_stmt(i.kid[0], arch)
        elif kid[0].name == "graph_block_spawn":
            kid = kid[0].kid[2].kid
            for i in kid:
                self.run_can_stmt(i, arch)
        self.pop_scope()

    # Note: Sentinels only registers the attr_stmts
    def load_walker(self, jac_ast):
        """
        walker: KW_WALKER NAME namespaces? walker_block;
        """
        kid = self.set_cur_ast(jac_ast)
        name = kid[1].token_text()
        kind = kid[0].token_text()
        walk = walker(m_id=self._m_id, h=self._h, code_ir=jac_ast, name=name, kind=kind)
        if jac_ast.kid[2].name == "namespaces":
            walk.namespaces = self.run_namespaces(jac_ast.kid[2])
        if self.walker_ids.has_obj_by_name(walk.name):
            self.walker_ids.destroy_obj_by_name(walk.name)
        self.walker_ids.add_obj(walk)
        return walk

    def load_test(self, jac_ast):
        """
        test:
            KW_TEST STRING KW_WITH (graph_ref | KW_GRAPH graph_block) KW_BY (
                (walker_ref spawn_ctx? (code_block | SEMI))
                | KW_WALKER walker_block
            );
        """
        kid = self.set_cur_ast(jac_ast)
        testcase = {
            "title": kid[1].token_text(),
            "graph_ref": None,
            "graph_block": None,
            "walker_ref": None,
            "spawn_ctx": None,
            "assert_block": None,
            "walker_block": None,
            "outcome": None,
        }
        kid = kid[3:]
        if kid[0].name == "graph_ref":
            graph_name = kid[0].kid[2].token_text()
            if not self.arch_ids.has_obj_by_name(graph_name, kind="graph"):
                self.rt_error(f"Graph {graph_name} not found!", kid[0])
                return
            testcase["graph_ref"] = graph_name
        else:
            kid = kid[1:]
            testcase["graph_block"] = jac_ast_to_ir(kid[0])
        kid = kid[2:]
        if kid[0].name == "walker_ref":
            walker_name = kid[0].kid[2].token_text()
            if not self.walker_ids.has_obj_by_name(walker_name):
                self.rt_error(f"Walker {walker_name} not found!", kid[0])
                return
            testcase["walker_ref"] = walker_name
            kid = kid[1:]
            if kid[0].name == "spawn_ctx":
                testcase["spawn_ctx"] = jac_ast_to_ir(kid[0])
                kid = kid[1:]
            if kid[0].name == "code_block":
                testcase["assert_block"] = jac_ast_to_ir(kid[0])
        else:
            kid = kid[1:]
            testcase["walker_block"] = jac_ast_to_ir(kid[0])

        self.testcases.append(testcase)

    def run_namespaces(self, jac_ast):
        """
        namespaces: COLON name_list;
        """
        return self.run_name_list(jac_ast.kid[1])



"""
Walker code generator for jac code in AST form
"""
from jaseci.jac.code_gen.code_gen import code_gen
from jaseci.jac.machine.op_codes import op
from jaseci.jac.machine.builtin_refs import ref
from jaseci.graph.node import node
from jaseci.jac.jac_set import jac_set


class walker_code_gen(code_gen):
    """Jac walker code generator class"""

    def gen_walkter(self, jac_ast):
        """
        walker:
            KW_WALKER NAME LBRACE attr_stmt* walk_entry_block? (
                statement
                | walk_activity_block
            )* walk_exit_block? RBRACE;
        """
        kid = jac_ast.kid
        self.g_ins([op.PUSH_SCOPE_W])
        self.g_ins([op.SET_LIVE_VAR, "here", ref.HERE_ID, []])
        l_skip_attr = self.next_lab()
        self.g_ins([op.B_NEQI, ref.STEP, 0, l_skip_attr])
        for i in kid:
            if i.name == "attr_stmt":
                self.gen_attr_stmt(jac_ast=i, obj=ref.SELF)
        self.g_lab(l_skip_attr)
        for i in kid:
            if i.name == "walk_entry_block":
                self.gen_walk_entry_block(i)
            if i.name == "statement":
                self.gen_statement(i)
            if i.name == "walk_activity_block":
                self.gen_walk_activity_block(i)
            if i.name == "walk_exit_block":
                self.gen_walk_exit_block(i)
        self.g_ins([op.POP_SCOPE])
        self.g_ins([op.END])

    def gen_walk_entry_block(self, jac_ast):
        """
        walk_entry_block: KW_WITH KW_ENTRY code_block;
        """
        kid = jac_ast.kid
        l_skip = self.next_lab()
        self.g_ins([op.B_NEQI, ref.STEP, 0, l_skip])
        self.g_ins([op.SET_REF_VARI, ref.IN_ENT_EXIT, True])
        self.gen_code_block(kid[2])
        self.g_ins([op.SET_REF_VARI, ref.IN_ENT_EXIT, False])
        self.g_lab(l_skip)

    def gen_walk_activity_block(self, jac_ast):
        """
        walk_activity_block: KW_WITH KW_ACTIVITY code_block;
        """
        kid = jac_ast.kid
        self.gen_code_block(kid[2])

    def gen_walker_action(self, jac_ast):
        """
        walker_action:
            ignore_action
            | take_action
            | destroy_action
            | KW_DISENGAGE SEMI;
        """
        kid = jac_ast.kid
        if kid[0].name == "KW_DISENGAGE":
            self.g_ins([op.SET_REF_VARI, ref.STOPPED, "stop"])
            self.g_ins([op.IDS_CLEAR, ref.NEXT_N_IDS])
        else:
            expr_func = getattr(self, f"gen_{kid[0].name}")
            expr_func(kid[0])

    def gen_ignore_action(self, jac_ast):
        """
        ignore_action: KW_IGNORE expression SEMI;
        """
        kid = jac_ast.kid
        self.gen_expression(kid[1])
        l_out = self.next_lab()
        l_skip = self.next_lab()
        self.g_ins([op.B_NIT, ref.RESULT_OUT, node, l_skip])
        self.g_ins([op.IDS_ADD_OBJ, ref.IGNORE_N_IDS, ref.RESULT_OUT])
        self.g_ins([op.B_A, l_out])
        self.g_lab(l_skip)

        l_skip = self.next_lab()
        self.g_ins([op.B_NIT, ref.RESULT_OUT, jac_set, l_skip])
        self.g_ins([op.PLUS, ref.IGNORE_N_IDS, ref.IGNORE_N_IDS, ref.RESULT_OUT])
        self.g_ins([op.B_A, l_out])
        self.g_lab(l_skip)
        self.gen_rt_error("is not ignorable type (i.e., nodes)", ref.RESULT_OUT)
        self.g_lab(l_out)

    def gen_take_action(self, jac_ast):
        """
        take_action:
            KW_TAKE expression (SEMI | else_stmt);
        """
        kid = jac_ast.kid
        self.gen_expression(kid[1])
        self.g_ins([op.IDS_GET_LEN, ref.TMP1, ref.NEXT_N_IDS])
        # before = len(self.next_node_ids)
        l_out = self.next_lab()
        l_skip = self.next_lab()
        self.g_ins([op.B_NIT, ref.RESULT_OUT, node, l_skip])
        self.g_ins([op.IDS_ADD_OBJ, ref.NEXT_N_IDS, ref.RESULT_OUT])
        self.g_ins([op.B_A, l_out])
        self.g_lab(l_skip)

        l_skip = self.next_lab()
        self.g_ins([op.B_NIT, ref.RESULT_OUT, jac_set, l_skip])
        self.g_ins([op.PLUS, ref.NEXT_N_IDS, ref.NEXT_N_IDS, ref.RESULT_OUT])
        self.g_ins([op.B_A, l_out])
        self.g_lab(l_skip)
        self.gen_rt_error("is not destination type (i.e., nodes)", ref.RESULT_OUT)
        self.g_lab(l_out)
        self.g_ins([op.IDS_GET_LEN, ref.TMP2, ref.NEXT_N_IDS])
        if kid[2].name == "else_stmt":
            l_skip = self.next_lab()
            self.g_ins([op.B_NEQ, ref.TMP1, ref.TMP2, l_skip])
            self.gen_else_stmt(kid[2])
            self.g_lab(l_skip)

    def gen_destroy_action(self, jac_ast):
        """
        destroy_action: KW_DESTROY expression SEMI;
        """
        kid = jac_ast.kid
        self.gen_expression(kid[1])
        l_out = self.next_lab()
        l_skip = self.next_lab()
        self.g_ins([op.B_NIT, ref.RESULT_OUT, node, l_skip])
        self.g_ins([op.IDS_ADD_OBJ, ref.DESTROY_N_IDS, ref.RESULT_OUT])
        self.g_ins([op.B_A, l_out])
        self.g_lab(l_skip)

        l_skip = self.next_lab()
        self.g_ins([op.B_NIT, ref.RESULT_OUT, jac_set, l_skip])
        self.g_ins([op.PLUS, ref.DESTROY_N_IDS, ref.DESTROY_N_IDS, ref.RESULT_OUT])
        self.g_ins([op.B_A, l_out])
        self.g_lab(l_skip)
        self.gen_rt_error("is not destroyable type (i.e., nodes)", ref.RESULT_OUT)
        self.g_lab(l_out)



from jaseci.utils.utils import TestCaseHelper
from unittest import TestCase

from jaseci.graph.node import node
from jaseci.element.master import master
from jaseci.element.super_master import super_master
from jaseci.element.element import element
from jaseci.graph.graph import graph
from jaseci.actor.sentinel import sentinel
from jaseci.utils.mem_hook import mem_hook
from jaseci.utils.utils import get_all_subclasses
import jaseci.tests.jac_test_code as jtc
import uuid


class architype_tests(TestCaseHelper, TestCase):
    def setUp(self):
        super().setUp()

    def tearDown(self):
        super().tearDown()

    def test_object_creation_basic_no_side_creation(self):
        """ """
        mast = master(h=mem_hook())
        num_objs = len(mast._h.mem.keys())
        node1 = node(m_id=mast._m_id, h=mast._h)
        node2 = node(m_id=mast._m_id, h=mast._h, parent_id=node1.id)
        num_new = len(mast._h.mem.keys())
        self.assertEqual(num_new, num_objs + 2)

        new_graph = graph(m_id=mast._m_id, h=mast._h)
        mast.graph_ids.add_obj(new_graph)
        num_new = len(mast._h.mem.keys())
        self.assertEqual(num_new, num_objs + 3)

        new_graph.attach_outbound(node1)
        new_graph.attach_outbound(node2)
        num_new = len(mast._h.mem.keys())
        self.assertEqual(num_new, num_objs + 5)

    def test_edge_removal_updates_nodes_edgelist(self):
        """ """
        mast = master(h=mem_hook())
        node1 = node(m_id=mast._m_id, h=mast._h)
        node2 = node(m_id=mast._m_id, h=mast._h)
        edge = node1.attach_outbound(node2)
        self.assertEqual(len(node1.edge_ids), 1)
        self.assertEqual(len(node2.edge_ids), 1)
        self.assertEqual(len(edge), 1)
        edge[0].destroy()
        self.assertEqual(len(node1.edge_ids), 0)
        self.assertEqual(len(node2.edge_ids), 0)

    def test_object_creation_by_sentinel_no_leaks(self):
        """
        Test that the destroy of sentinels clears owned objects
        """
        mast = master(h=mem_hook())
        num_objs = len(mast._h.mem.keys()) - len(mast._h.global_action_list)
        self.assertEqual(num_objs, 2)
        new_graph = graph(m_id=mast._m_id, h=mast._h)
        sent = sentinel(m_id=mast._m_id, h=mast._h)
        code = jtc.prog1
        mast.sentinel_ids.add_obj(sent)
        mast.graph_ids.add_obj(new_graph)
        num_new = len(mast._h.mem.keys()) - len(mast._h.global_action_list)
        self.assertEqual(num_new, num_objs + 2 + 3)

        sent.register_code(code)
        num_objs = len(mast._h.mem.keys()) - len(mast._h.global_action_list)
        sent.register_code(code)
        new_num = len(mast._h.mem.keys()) - len(mast._h.global_action_list)
        self.assertEqual(num_objs, new_num)

    def test_json_blob_of_objects(self):
        """
        Test saving object to json and back to python dict
        """
        for i in get_all_subclasses(element):
            orig = i(m_id="anon", h=mem_hook())
            blob1 = orig.json(detailed=True)
            new = i(m_id="anon", h=mem_hook())
            self.assertNotEqual(orig.id, new.id)
            new.json_load(blob1)
            self.assertEqual(orig.id, new.id)
            self.assertTrue(orig.is_equivalent(new))

    def test_supermaster_can_touch_all_data(self):
        mh = mem_hook()
        mast = master(h=mh)
        mast2 = master(h=mh)
        node12 = node(m_id=mast2._m_id, h=mast2._h)
        supmast = super_master(h=mh)
        bad = mh.get_obj(mast._m_id, uuid.UUID(node12.jid))
        good = mh.get_obj(supmast._m_id, uuid.UUID(node12.jid))
        self.assertEqual(good, node12)
        self.assertNotEqual(bad, node12)
        self.assertIsNone(bad)



"""
Interpreter for jac code in AST form

This interpreter should be inhereted from the class that manages state
referenced through self.
"""
from jaseci.utils.utils import logger
from jaseci.actions.live_actions import live_actions, load_preconfig_actions

# from jaseci.actions.find_action import find_action
from jaseci.element.element import element

from jaseci.jac.jac_set import jac_set
from jaseci.jac.machine.jac_scope import jac_scope


class machine_state:
    """Shared interpreter class across both sentinels and walkers"""

    def __init__(self, parent_override=None, caller=None):
        self.report = []
        self.report_status = None
        self.runtime_errors = []
        self._parent_override = parent_override
        if not isinstance(self, element) and caller:
            self._m_id = caller._m_id
            self._h = caller._h
        self._scope_stack = [None]
        self._jac_scope = None
        self._relevant_edges = []
        self._loop_ctrl = None
        self._stopped = None
        self._assign_mode = False
        self._loop_limit = 10000
        self._cur_jac_ast = None

    def parent(self):
        if self._parent_override:
            return self._parent_override
        else:
            return element.parent(self)

    def reset(self):
        self.report = []
        self.report_status = None
        self.runtime_errors = []
        self._scope_stack = [None]
        self._jac_scope = None
        self._loop_ctrl = None
        self._stopped = None

    def push_scope(self, scope: jac_scope):
        self._scope_stack.append(scope)
        self._jac_scope = scope

    def pop_scope(self):
        self._scope_stack.pop()
        self._jac_scope = self._scope_stack[-1]

    def set_cur_ast(self, jac_ast):
        self._cur_jac_ast = jac_ast
        return jac_ast.kid

    # Helper Functions ##################

    def get_arch_for(self, obj):
        """Returns the architype that matches object"""
        ret = self.parent().arch_ids.get_obj_by_name(name=obj.name, kind=obj.kind)
        if ret is None:
            self.rt_error(f"Unable to find architype for {obj.name}, {obj.kind}")
        return ret

    def obj_set_to_jac_set(self, obj_set):
        """
        Returns nodes jac_set from edge jac_set from current node
        """
        ret = jac_set()
        for i in obj_set:
            ret.add_obj(i)
        return ret

    def edge_to_node_jac_set(self, edge_set):
        """
        Returns nodes jac_set from edge jac_set from current node
        """
        ret = jac_set()
        for i in edge_set.obj_list():
            ret.add_obj(i.opposing_node(self.current_node))
        return ret

    def edges_filter_on_nodes(self, edge_set, node_set):
        """
        Returns nodes jac_set from edge jac_set from current node
        """
        ret = jac_set()
        for i in edge_set.obj_list():
            for j in node_set.obj_list():
                if i.jid in j.edge_ids:
                    ret.add_obj(i)
                    break
        return ret

    def get_builtin_action(self, func_name, jac_ast=None):
        """
        Takes reference to action attr, finds the built in function
        and returns new name used as hook by action class
        """
        if func_name not in live_actions.keys():
            load_preconfig_actions(self._h)
        if func_name not in live_actions.keys():
            self.rt_error(f"Builtin action not found - {func_name}", jac_ast)
        return func_name

    def jac_exception(self, e: Exception, jac_ast):
        return {
            "type": type(e).__name__,
            "mod": jac_ast.mod_name,
            "msg": str(e),
            "args": e.args,
            "line": jac_ast.line,
            "col": jac_ast.column,
        }

    def rt_log_str(self, msg, jac_ast=None):
        """Generates string for screen output"""
        if jac_ast is None:
            jac_ast = self._cur_jac_ast
        name = self.name if hasattr(self, "name") else "blank"
        if jac_ast:
            msg = (
                f"{jac_ast.mod_name}:{name} - line {jac_ast.line}, "
                + f"col {jac_ast.column} - rule {jac_ast.name} - {msg}"
            )
        else:
            msg = f"{msg}"
        return msg

    def rt_warn(self, error, jac_ast=None):
        """Prints runtime error to screen"""
        error = self.rt_log_str(error, jac_ast)
        logger.warning(str(error))
        self.runtime_errors.append(error)

    def rt_error(self, error, jac_ast=None):
        """Prints runtime error to screen"""
        error = self.rt_log_str(error, jac_ast)
        logger.error(str(error))
        self.runtime_errors.append(error)

    def rt_info(self, msg, jac_ast=None):
        """Prints runtime info to screen"""
        logger.info(str(self.rt_log_str(msg, jac_ast)))

    def rt_check_type(self, obj, typ, jac_ast=None):
        """Prints error if type mismatach"""
        if not isinstance(typ, list):
            typ = [typ]
        for i in typ:
            if isinstance(obj, i):
                return True
        self.rt_error(
            f"Incompatible type for object "
            f"{obj} - {type(obj).__name__}, "
            f"expecting {typ}",
            jac_ast,
        )
        return False



"""
Master api as a mixin
"""
from jaseci.api.interface import interface
from jaseci.utils.id_list import id_list


class master_api:
    """Master APIs for creating nicknames for UUIDs and other long strings"""

    def __init__(self, head_master):
        self._caller = self
        self.head_master_id = head_master
        self.sub_master_ids = id_list(self)

    @interface.private_api(cli_args=["name"])
    def master_create(
        self, name: str, set_active: bool = True, other_fields: dict = {}
    ):
        """
        Create a master instance and return root node master object

        other_fields used for additional feilds for overloaded interfaces
        (i.e., Dango interface)
        """
        from jaseci.element.master import master

        new_m = master(h=self._h, name=name)
        return self.make_me_head_master_or_destroy(new_m)

    @interface.private_api(cli_args=["name"])
    def master_get(self, name: str, mode: str = "default", detailed: bool = False):
        """
        Return the content of the master with mode
        Valid modes: {default, }
        """
        mas = self.sub_master_ids.get_obj_by_name(name)
        if not mas:
            return {"response": f"{name} not found"}
        else:
            return mas.serialize(detailed=detailed)

    @interface.private_api()
    def master_list(self, detailed: bool = False):
        """
        Provide complete list of all master objects (list of root node objects)
        """
        masts = []
        for i in self.sub_master_ids.obj_list():
            masts.append(i.serialize(detailed=detailed))
        return masts

    @interface.private_api(cli_args=["name"])
    def master_active_set(self, name: str):
        """
        Sets the default master master should use
        NOTE: Specail handler included in general_interface_to_api
        """
        mas = self.sub_master_ids.get_obj_by_name(name)
        if not mas:
            return {"response": f"{name} not found"}
        self._caller = mas
        return {"response": f"You are now {mas.name}"}

    @interface.private_api()
    def master_active_unset(self):
        """
        Unsets the default sentinel master should use
        """
        self._caller = self
        return {"response": f"You are now {self.name}"}

    @interface.private_api()
    def master_active_get(self, detailed: bool = False):
        """
        Returns the default master master is using
        """
        return self._caller.serialize(detailed=detailed)

    @interface.private_api()
    def master_self(self, detailed: bool = False):
        """
        Returns the masters object
        """
        return self.serialize(detailed=detailed)

    @interface.private_api(cli_args=["name"])
    def master_delete(self, name: str):
        """
        Permanently delete master with given id
        """
        if not self.sub_master_ids.has_obj_by_name(name):
            return {"response": f"{name} not found"}
        self.sub_master_ids.destroy_obj_by_name(name)
        return {"response": f"{name} has been destroyed"}

    def make_me_head_master_or_destroy(self, m):
        """
        Utility to bring an object into sub masters
        """
        m.head_master_id = self.jid
        m.give_access(self)
        if self.sub_master_ids.has_obj_by_name(m.name):
            name = m.name
            m.destroy()
            return {"response": f"{name} already exists"}
        m.save()
        self.sub_master_ids.add_obj(m)
        return m.serialize()

    def destroy(self):
        """
        Destroys self from memory and persistent storage
        """
        for i in self.sub_master_ids.obj_list():
            i.destroy()



from jaseci.utils.utils import logger
import jaseci.actions.standard.date as jsdate
from jaseci.actions.live_actions import jaseci_action


@jaseci_action()
def year_from_date(date: str):
    logger.warning("Deprecated! Use date.quantlize_...")
    return jsdate.quantize_to_year(date)


@jaseci_action()
def month_from_date(date: str):
    logger.warning("Deprecated! Use date.quantlize_...")
    return jsdate.quantize_to_month(date)


@jaseci_action()
def week_from_date(date: str):
    logger.warning("Deprecated! Use date.quantlize_...")
    return jsdate.quantize_to_week(date)


@jaseci_action()
def day_from_date(date: str):
    logger.warning("Deprecated! Use date.quantlize_...")
    return jsdate.quantize_to_day(date)


@jaseci_action()
def date_day_diff(start_date: str, end_date: str):
    logger.warning("Deprecated! Use date...")
    return jsdate.date_day_diff(start_date, end_date)



"""
Walker api functions as a mixin
"""
from jaseci.api.interface import interface
from jaseci.actor.walker import walker
from jaseci.graph.node import node
from jaseci.actor.sentinel import sentinel
from jaseci.utils.utils import b64decode_str
from jaseci.utils.id_list import id_list


class walker_api:
    """
    Walker APIs
    """

    def __init__(self):
        self.spawned_walker_ids = id_list(self)

    @interface.public_api(cli_args=["wlk"])
    def walker_summon(
        self, key: str, wlk: walker, nd: node, ctx: dict = {}, global_sync: bool = True
    ):
        """
        Public api for running walkers, namespace key must be provided
        along with the walker id and node id
        """
        if key not in wlk.namespace_keys().values():
            return self.bad_walk_response(["Not authorized to execute this walker"])
        if global_sync:  # Test needed
            self.sync_walker_from_global_sent(wlk)
        walk = wlk.duplicate()
        walk.refresh()
        walk.prime(nd, prime_ctx=ctx)
        res = walk.run()
        walk.destroy()
        return res

    @interface.private_api(cli_args=["code"])
    def walker_register(
        self, snt: sentinel = None, code: str = "", encoded: bool = False
    ):
        """
        Create blank or code loaded walker and return object
        """
        if encoded:
            code = b64decode_str(code)
        walk = snt.register_walker(code)
        if walk:
            self.extract_wlk_aliases(snt, walk)
            return walk.serialize()
        else:
            return ["Walker not created, invalid code!"]

    @interface.private_api(cli_args=["wlk"])
    def walker_get(self, wlk: walker, mode: str = "default", detailed: bool = False):
        """
        Get a walker rendered with specific mode
        Valid modes: {default, code, ir, keys, }
        """
        if mode == "code":
            return wlk._jac_ast.get_text()
        elif mode == "ir":
            return wlk.ir_dict()
        elif mode == "keys":
            return wlk.namespace_keys()
        else:
            return wlk.serialize(detailed=detailed)

    @interface.private_api(cli_args=["wlk"])
    def walker_set(self, wlk: walker, code: str, mode: str = "default"):
        """
        Set code/ir for a walker
        Valid modes: {code, ir, }
        """
        if mode == "code" or mode == "default":
            wlk.register(code)
        elif mode == "ir":
            wlk.apply_ir(code)
        else:
            return [f"Invalid mode to set {wlk}"]
        if wlk.is_active:
            return [f"{wlk} registered and active!"]
        else:
            return [f"{wlk} code issues encountered!"]

    @interface.private_api()
    def walker_list(self, snt: sentinel = None, detailed: bool = False):
        """
        List walkers known to sentinel
        """
        walks = []
        for i in snt.walker_ids.obj_list():
            walks.append(i.serialize(detailed=detailed))
        return walks

    @interface.private_api(cli_args=["wlk"])
    def walker_delete(self, wlk: walker, snt: sentinel = None):
        """
        Permanently delete walker with given id
        """
        self.remove_wlk_aliases(snt, wlk)
        wlkid = wlk.id
        snt.walker_ids.destroy_obj(wlk)
        return [f"Walker {wlkid} successfully deleted"]

    @interface.private_api(cli_args=["name"])
    def walker_spawn_create(self, name: str, snt: sentinel = None):
        """
        Creates new instance of walker and returns new walker object
        """
        wlk = snt.spawn_walker(name, caller=self)
        if wlk:
            if self.spawned_walker_ids.has_obj_by_name(name):
                self.spawned_walker_ids.destroy_obj_by_name(name)
            self.spawned_walker_ids.add_obj(wlk)
            self.alias_register(f"spawned:walker:{name}", wlk.jid)
            return wlk.serialize()
        else:
            return ["Walker not found!"]

    @interface.private_api(cli_args=["name"])
    def walker_spawn_delete(self, name: str):
        """
        Delete instance of walker
        """
        if self.spawned_walker_ids.has_obj_by_name(name):
            self.spawned_walker_ids.destroy_obj_by_name(name)
            self.alias_delete(f"spawned:walker:{name}")
            return [f"Walker {name} deteled!"]
        else:
            return [f"Walker {name} not found!"]

    @interface.private_api()
    def walker_spawn_list(self, detailed: bool = False):
        """
        List walkers spawned by master
        """
        walks = []
        for i in self.spawned_walker_ids.obj_list():
            walks.append(i.serialize(detailed=detailed))
        return walks

    @interface.private_api(cli_args=["wlk"])
    def walker_prime(self, wlk: walker, nd: node = None, ctx: dict = {}):
        """
        Assigns walker to a graph node and primes walker for execution
        """
        wlk.prime(nd, prime_ctx=ctx)
        return [f"Walker primed on node {nd.id}"]

    @interface.private_api(cli_args=["wlk"])
    def walker_execute(
        self, wlk: walker, prime: node = None, ctx: dict = {}, profiling: bool = False
    ):
        """
        Executes walker (assumes walker is primed)
        """
        if prime:
            self.walker_prime(wlk=wlk, nd=prime, ctx=ctx)
        return wlk.run(profiling=profiling)

    @interface.private_api(cli_args=["name"])
    def walker_run(
        self,
        name: str,
        nd: node = None,
        ctx: dict = {},
        snt: sentinel = None,
        profiling: bool = False,
    ):
        """
        Creates walker instance, primes walker on node, executes walker,
        reports results, and cleans up walker instance.
        """
        wlk = snt.spawn_walker(name, caller=self)
        if not wlk:
            return self.bad_walk_response([f"Walker {name} not found!"])
        res = self.walker_execute(wlk=wlk, prime=nd, ctx=ctx, profiling=profiling)
        wlk.destroy()
        return res

    @interface.private_api(cli_args=["name"], url_args=["name"])
    def wapi(
        self,
        name: str,
        nd: node = None,
        ctx: dict = {},
        snt: sentinel = None,
        profiling: bool = False,
    ):
        """
        Walker individual APIs
        """
        return self.walker_run(name, nd, ctx, snt, profiling)

    def destroy(self):
        """
        Destroys self from memory and persistent storage
        """
        for i in self.spawned_walker_ids.obj_list():
            i.destroy()

    def bad_walk_response(self, errors=list()):
        return {"report": [], "success": False, "errors": errors}



"""
Super (master) api as a mixin
"""
from jaseci.api.interface import interface
from jaseci.element.master import master


class super_api:
    """Super APIs for creating nicknames for UUIDs and other long strings"""

    @interface.admin_api(cli_args=["name"])
    def master_createsuper(
        self, name: str, set_active: bool = True, other_fields: dict = {}
    ):
        """
        Create a super instance and return root node super object

        other_fields used for additional feilds for overloaded interfaces
        (i.e., Dango interface)
        """
        from jaseci.element.super_master import super_master

        new_m = super_master(h=self._h, name=name)
        return self.make_me_head_master_or_destroy(new_m)

    @interface.admin_api()
    def master_allusers(self, num: int = 0, start_idx: int = 0):
        """
        Returns info on a set of users, num specifies the number of users to
        return and start idx specfies where to start
        NOTE: Abstract interface to be overridden
        """

    @interface.admin_api(cli_args=["mast"])
    def master_become(self, mast: master):
        """
        Sets the default master master should use
        FIXME: _caller does not persist accross http request!!
        """
        self._caller = mast
        return {"response": f"You are now {mast.name}"}



"""
General master interface engine for client interfaces as mixin
"""
import uuid
from inspect import signature, getdoc
from jaseci.utils.utils import logger
from jaseci.utils.utils import is_jsonable
from jaseci.element.element import element


class interface:
    """
    General master interface engine
    """

    _public_api = []
    _private_api = []
    _admin_api = []
    _cli_api = []

    def __init__(self):
        """
        self._pub_committer is set by api implementaiton if intent
        is to commit changes enacted by public call
        """
        self._pub_committer = None

    def assimilate_api(api_list, func, cmd_group=None, cli_args=None, url_args=None):
        cmd_group = func.__name__.split("_") if cmd_group is None else cmd_group
        api_list.append(
            {
                "fname": func.__name__,
                "sig": signature(func),
                "doc": getdoc(func),
                "groups": cmd_group,
                "cli_args": cli_args if cli_args is not None else [],
                "url_args": url_args if url_args is not None else [],
            }
        )
        return func

    def public_api(cmd_group=None, cli_args=None, url_args=None):
        def decorator_func(func):
            return interface.assimilate_api(
                interface._public_api, func, cmd_group, cli_args, url_args
            )

        return decorator_func

    def private_api(cmd_group=None, cli_args=None, url_args=None):
        def decorator_func(func):
            return interface.assimilate_api(
                interface._private_api, func, cmd_group, cli_args, url_args
            )

        return decorator_func

    def admin_api(cmd_group=None, cli_args=None, url_args=None):
        def decorator_func(func):
            return interface.assimilate_api(
                interface._admin_api, func, cmd_group, cli_args, url_args
            )

        return decorator_func

    def cli_api(cmd_group=None, cli_args=None):
        def decorator_func(func):
            return interface.assimilate_api(
                interface._cli_api, func, cmd_group, cli_args
            )

        return decorator_func

    def all_apis(self):
        return interface._public_api + interface._private_api + interface._admin_api

    assimilate_api = staticmethod(assimilate_api)
    public_api = staticmethod(public_api)
    private_api = staticmethod(private_api)
    admin_api = staticmethod(admin_api)
    cli_api = staticmethod(cli_api)

    def provide_internal_default(self, param):
        """
        Applies internal defaults for sentinel and graphs
        """
        if param == "snt" and self.active_snt_id:
            if self.active_snt_id == "global":
                glob_id = self._h.get_glob("GLOB_SENTINEL")
                if not glob_id:
                    return self.interface_error("No global sentinel is available!")
                else:
                    return glob_id
            return self.active_snt_id
        if param == "gph" or param == "nd":
            return (
                self.active_gph_id
                if self.active_gph_id
                else self.interface_error("No default graph node available!")
            )
        return None

    def interface_error(self, err, stack=None):
        """Standard error output to logger and api response"""
        logger.error(err)
        ret = {"response": err, "success": False, "errors": [err]}
        if stack:
            ret["stack_trace"] = stack
        return ret

    def general_interface_to_api(self, params, api_name):
        """
        A mapper utility to interface to master class
        Assumptions:
            params is a dictionary of parameter names and values in UUID
            api_name is the name of the api being mapped to
        """
        param_map = {}
        if api_name.startswith("master_active"):
            _caller = self
        else:
            _caller = self._caller
        if not hasattr(_caller, api_name):
            return self.interface_error(f"{api_name} not a valid API")
        func_sig = signature(getattr(_caller, api_name))
        for i in func_sig.parameters.keys():
            if i == "self":
                continue
            p_name = i
            p_type = func_sig.parameters[i].annotation
            p_default = func_sig.parameters[i].default
            val = p_default if p_default is not func_sig.parameters[i].empty else None
            if p_name in params.keys():
                val = params[p_name]
            if val is None:  # Used to patch defaults
                val = _caller.provide_internal_default(p_name)
                if val is not None and "errors" in val:
                    return val
            if str(val) in _caller.alias_map.keys():
                val = _caller.alias_map[val]
            if issubclass(p_type, element):
                if val is None:
                    break
                val = _caller._h.get_obj(_caller._m_id, uuid.UUID(val))
                if isinstance(val, p_type):
                    param_map[i] = val
                else:
                    return self.interface_error(f"{type(val)} is not {p_type}")
            else:  # TODO: Can do type checks here too
                param_map[i] = val

            if param_map[i] is None:
                return self.interface_error(f"Invalid API args - {params}")
        try:
            ret = getattr(_caller, api_name)(**param_map)
        except Exception as e:
            import traceback as tb

            return self.interface_error(
                f"Internal Exception: {e}", stack=tb.format_exc()
            )
        if not is_jsonable(ret):
            return self.interface_error(f"Non-JSON API ret {type(ret)}: {ret}")
        return ret

    def public_interface_to_api(self, params, api_name):
        """
        A mapper utility to interface to public
        Assumptions:
            params is a dictionary of parameter names and values in UUID
            api_name is the name of the api being mapped to
        """
        param_map = {}
        if not hasattr(self, api_name):
            return self.interface_error(f"{api_name} not a valid API")
            return False
        func_sig = signature(getattr(self, api_name))
        for i in func_sig.parameters.keys():
            if i == "self":
                continue
            p_name = i
            p_type = func_sig.parameters[i].annotation
            p_default = func_sig.parameters[i].default
            val = p_default if p_default is not func_sig.parameters[i].empty else None
            if p_name in params.keys():
                val = params[p_name]
            if issubclass(p_type, element):
                if val is None:
                    return self.interface_error(
                        f"No {p_type} value for {p_name} provided!"
                    )
                val = self._h.get_obj("override", uuid.UUID(val), override=True)
                self.seek_committer(val)
                if isinstance(val, p_type):
                    param_map[i] = val
                else:
                    return self.interface_error(f"{type(val)} is not {p_type}")
                    param_map[i] = None
            else:  # TODO: Can do type checks here too
                param_map[i] = val

            if param_map[i] is None:
                return self.interface_error(f"Invalid API parameter set - {params}")
        try:
            ret = getattr(self, api_name)(**param_map)
        except Exception as e:
            import traceback as tb

            return self.interface_error(
                f"Internal Exception: {e}", stack=tb.format_exc()
            )
        if not is_jsonable(ret):
            return self.interface_error(
                str(f"API returns non json object {type(ret)}: {ret}")
            )
        return ret

    def seek_committer(self, obj):
        """Opportunistically assign a committer"""
        if not self._pub_committer:
            self._pub_committer = obj._h.get_obj(obj._m_id, uuid.UUID(obj._m_id))

    def clear_committer(self):
        """Unset committer"""
        self._pub_committer = None

    def sync_walker_from_global_sent(self, wlk):
        """Checks for matching code ir between global and spawned walker"""
        glob_id = wlk._h.get_glob("GLOB_SENTINEL")
        if glob_id:
            snt = wlk._h.get_obj(wlk._m_id, uuid.UUID(glob_id))
            glob_wlk = snt.walker_ids.get_obj_by_name(wlk.name)
            if glob_wlk and glob_wlk.code_sig != wlk.code_sig:
                wlk.apply_ir(glob_wlk.code_ir)



from jaseci.utils.mem_hook import mem_hook
from jaseci.actor.sentinel import sentinel
from jaseci.graph.graph import graph
from jaseci.element.super_master import super_master
from jaseci.element.master import master
from jaseci.utils.utils import TestCaseHelper
from unittest import TestCase
import jaseci.tests.jac_test_progs as jtp


class jac_tests(TestCaseHelper, TestCase):
    """Unit tests for Jac language"""

    def setUp(self):
        super().setUp()

    def tearDown(self):
        super().tearDown()

    def test_bug_check1(self):
        gph = graph(m_id="anon", h=mem_hook())
        sent = sentinel(m_id="anon", h=gph._h)
        sent.register_code(jtp.bug_check1)
        test_walker = sent.walker_ids.get_obj_by_name("init")
        test_walker.prime(gph)
        test_walker.run()
        report = test_walker.report
        self.assertEqual(report[0][0], "THIS IS AN INTENT_LABEL")

    def test_action_load_std_lib(self):
        mast = super_master(h=mem_hook())
        mast.sentinel_register(name="test", code=jtp.action_load_std_lib)
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "aload"}
        )["report"]
        self.assertEqual(report[0], True)

    def test_action_load_std_lib_only_super(self):
        mast = master(h=mem_hook())
        mast.sentinel_register(name="test", code=jtp.action_load_std_lib)
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "aload"}
        )
        report = report["report"]
        self.assertEqual(report[0], False)

    def test_globals(self):
        gph = graph(m_id="anon", h=mem_hook())
        sent = sentinel(m_id="anon", h=gph._h)
        sent.register_code(jtp.globals)
        test_walker = sent.walker_ids.get_obj_by_name("init")
        test_walker.prime(gph)
        test_walker.run()
        report = test_walker.report
        self.assertEqual(report, ["testing", 56])

    def test_net_root_std_lib(self):
        mast = master(h=mem_hook())
        mast.sentinel_register(name="test", code=jtp.net_root_std_lib)
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )["report"]
        self.assertEqual(report[0][0], report[0][1])
        self.assertEqual(report[0][1], report[1][1])
        self.assertNotEqual(report[1][0], report[1][1])

    def test_or_stmt(self):
        mast = master(h=mem_hook())
        mast.sentinel_register(name="test", code=jtp.or_stmt)
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )["report"]
        self.assertEqual(report, [[3.4, "Hello"]])

    def test_nd_equals(self):
        mast = master(h=mem_hook())
        mast.sentinel_register(name="test", code=jtp.nd_equals_error_correct_line)
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )
        self.assertIn("line 3", report["errors"][0])

    def test_strange_ability_bug(self):
        mast = master(h=mem_hook())
        mast.sentinel_register(name="test", code=jtp.strange_ability_bug)
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "travel"}
        )["report"]
        mast.sentinel_register(name="test", code=jtp.strange_ability_bug, auto_run="")
        report += mast.general_interface_to_api(
            api_name="walker_run", params={"name": "travel"}
        )["report"]
        ir = mast.sentinel_get(mode="ir", snt=mast.active_snt())
        mast.sentinel_set(code=ir, snt=mast.active_snt(), mode="ir")
        report += mast.general_interface_to_api(
            api_name="walker_run", params={"name": "travel"}
        )["report"]
        report += mast.general_interface_to_api(
            api_name="walker_run", params={"name": "travel"}
        )["report"]
        self.assertEqual(report, ["Showing", "Showing", "Showing", "Showing"])

    def test_node_inheritance(self):
        mast = master(h=mem_hook())
        mast.sentinel_register(name="test", code=jtp.node_inheritance, auto_run="")
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )
        self.assertEqual(
            report,
            {
                "report": [
                    "plain.x",
                    {"a": 55, "b": 7, "c": 7, "d": 80},
                    "super.x",
                    "plain2.y",
                    "super.y",
                ],
                "success": True,
            },
        )

    def test_node_inheritance_chain_check(self):
        mast = master(h=mem_hook())
        mast.sentinel_register(
            name="test", code=jtp.node_inheritance_chain_check, auto_run=""
        )
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )
        self.assertEqual(report["success"], False)

    def test_global_reregistering(self):
        mast = master(h=mem_hook())
        mast.sentinel_register(name="test", code=jtp.global_reregistering)
        self.assertTrue(mast.active_snt().is_active)
        mast.sentinel_set(snt=mast.active_snt(), code=jtp.global_reregistering)
        self.assertTrue(mast.active_snt().is_active)
        mast.sentinel_register(name="test", code=jtp.global_reregistering)
        self.assertTrue(mast.active_snt().is_active)

    def test_vector_cos_sim_check(self):
        mast = master(h=mem_hook())
        mast.sentinel_register(name="test", code=jtp.vector_cos_sim_check, auto_run="")
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )["report"]
        self.assertEqual(len(report), 1)
        self.assertEqual(type(report[0]), float)

    def test_multi_breaks(self):
        mast = master(h=mem_hook())
        mast.sentinel_register(name="test", code=jtp.multi_breaks, auto_run="")
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )["report"]
        self.assertEqual(len(report), 15)
        self.assertEqual(report[14], 180)

    def test_reffy_deref_check(self):
        self.logger_on()
        mast = master(h=mem_hook())
        mast.sentinel_register(name="test", code=jtp.reffy_deref_check, auto_run="")
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )["report"]
        self.assertFalse(report[0])
        self.assertTrue(report[1])



from jaseci.utils.utils import obj_class_cache, build_class_dict

# from pprint import pformat
from inspect import getdoc
import jaseci


class book:
    def api_cheatsheet(self, root, out=None, str=""):
        if out is None:
            out = []
        if "leaf" in root.keys():
            out.append(str + f'& {root["leaf"][1]} &\n')
            return
        for i in root.keys():
            self.api_cheatsheet(root[i], out, str + f"{i} ")
        ret = ""
        for i in out:
            ret += i
        return ret

    def api_spec(self):
        ret = ""
        build_class_dict(jaseci)
        for i in obj_class_cache.keys():
            if not i.endswith("_api"):
                continue
            ret += f"\\subsection{{{i[:-4]} APIs}}\n\n"
            doc = getdoc(obj_class_cache[i]).replace("\n\n", "\n\\par\n")
            doc = doc.replace("_", "\\_")
            ret += f"{doc}\n\n"
            ret += self.api_call_spec(obj_class_cache[i])
        return ret

    def api_call_spec(self, cls):
        ret = ""
        for i in dir(cls):
            # access = 'master'
            if i.startswith("api_"):
                api = i[4:].replace("_", " ")
            elif i.startswith("admin_api_"):
                # access = 'super'
                api = i[10:].replace("_", " ")
            elif i.startswith("public_api_"):
                # access = 'public'
                api = i[11:].replace("_", " ")
            else:
                continue
            ret += f"\\subsubsection{{{api}}}\n\n"
        return ret



from setuptools import setup, find_packages

setup(
    name="jaseci",
    version="1.3.3.15",
    packages=find_packages(include=["jaseci", "jaseci.*"]),
    install_requires=[
        "click>=8.1.0,<8.2.0",
        "click-shell>=2.1,<2.2",
        "numpy >= 1.22.3, < 1.23.0",
        "antlr4-python3-runtime>=4.9.3,<4.10.0",
        "fastapi[all]>=0.75.0,<1.0.0",
        "requests",
        "redis",
        "flake8",
        "stripe",
        "pydantic",
    ],
    package_data={
        "": ["*.ini"],
    },
    entry_points={"console_scripts": ["jsctl = jaseci.jsctl.jsctl:main"]},
)



rand_std = """
    walker init
    {
        report rand.word();
        report rand.sentence();
        report rand.paragraph();
        report rand.text();
    }
    """

file_io = """
    walker init {
        fn="fileiotest.txt";
        a = {'a': 5};
        file.dump_json(fn, a);
        b=file.load_json(fn);
        b['a']+=b['a'];
        file.dump_json(fn, b);
        c=file.load_str(fn);
        file.append_str(fn, c);
        c=file.load_str(fn);
        report c;
        file.delete(fn);
    }
    """

std_used_in_node_has_var = """
    node testnode {has a=rand.sentence();}

    walker init
    {
        a = spawn here --> node::testnode;
        a = spawn here --> node::testnode;
        a = spawn --> --> node::testnode;
        report a[0].a;
    }
    """



from .id_list import id_list
from .utils import logger

import json


def json_str_to_jsci_dict(input_str, parent_obj=None):
    """
    Helper function to convert JSON strings to dictionarys with _ids list
    conversions from hex to UUID

    ret_obj is the owning object for id_list objects
    """
    try:
        obj_fields = json.loads(input_str)
    except ValueError:
        logger.error(str(f"Invalid jsci_obj string {input_str} on {parent_obj.id.urn}"))
        obj_fields = {}
    for i in obj_fields.keys():
        if str(i).endswith("_ids") and isinstance(obj_fields[i], list):
            obj_fields[i] = id_list(parent_obj=parent_obj, in_list=obj_fields[i])
    return obj_fields


class mem_hook:
    """
    Set of virtual functions to be used as hooks to allow access to
    the complete set of items across jaseci object types. This class contains
    a number of blank function calls to be bound extrenally and passed
    to the objects. They return jaseci core types.
    """

    def __init__(self):
        from jaseci.actions.live_actions import get_global_actions

        self.mem = {"global": {}}
        self.global_action_list = get_global_actions(self)

    def get_obj(self, caller_id, item_id, override=False):
        """
        Get item from session cache by id, then try store
        TODO: May need to make this an object copy so you cant do mem writes
        """
        if item_id in self.mem.keys():
            ret = self.mem[item_id]
            if override or ret.check_read_access(caller_id):
                return ret
        else:
            ret = self.get_obj_from_store(item_id)
            self.mem[item_id] = ret
            if override or (ret is not None and ret.check_read_access(caller_id)):
                return ret

    def has_obj(self, item_id):
        """
        Checks for object existance
        """
        if item_id in self.mem.keys():
            return True
        else:
            return self.has_obj_in_store(item_id)

    def save_obj(self, caller_id, item, persist=False):
        """Save item to session cache, then to store"""
        if item.check_write_access(caller_id):
            self.mem[item.id] = item
            if persist:
                self.save_obj_to_store(item)

    def destroy_obj(self, caller_id, item, persist=False):
        """Destroy item from session cache then  store"""
        if item.check_write_access(caller_id):
            self.mem[item.id] = None
            del self.mem[item.id]
            if persist:
                self.destroy_obj_from_store(item)

    def get_glob(self, name):
        """
        Get global config from session cache by id, then try store
        """
        if name in self.mem["global"].keys():
            return self.mem["global"][name]
        else:
            ret = self.get_glob_from_store(name)
            self.mem["global"][name] = ret
            return ret

    def has_glob(self, name):
        """
        Checks for global config existance
        """
        if name in self.mem["global"].keys():
            return True
        else:
            return self.has_glob_in_store(name)

    def resolve_glob(self, name, default=None):
        """
        Util function for returning config if exists otherwise default
        """
        if self.has_glob(name):
            return self.get_glob(name)
        else:
            return default

    def save_glob(self, name, value, persist=True):
        """Save global config to session cache, then to store"""
        self.mem["global"][name] = value
        if persist:
            self.save_glob_to_store(name, value)

    def list_glob(self):
        """Lists all configs present"""
        glob_list = self.list_glob_from_store()
        if glob_list:
            return glob_list
        else:
            return list(self.mem["global"].keys())

    def destroy_glob(self, name, persist=True):
        """Destroy global config from session cache then store"""
        self.mem["global"][name] = None
        del self.mem["global"][name]
        if persist:
            self.destroy_glob_from_store(name)

    def clear_mem_cache(self):
        """
        Clears memory, should only be used if underlying store is modified
        through other means than methods of this class
        """
        mem_hook.__init__(self)

    def get_obj_from_store(self, item_id):
        """
        Get item from externally hooked general store by id
        """

    def has_obj_in_store(self, item_id):
        """
        Checks for object existance in store
        """
        return False

    def save_obj_to_store(self, item):
        """Save item to externally hooked general store"""

    def destroy_obj_from_store(self, item):
        """Destroy item to externally hooked general store"""

    def get_glob_from_store(self, name):
        """
        Get global config from externally hooked general store by name
        """

    def has_glob_in_store(self, name):
        """
        Checks for global config existance in store
        """
        return False

    def save_glob_to_store(self, name, value):
        """Save global config to externally hooked general store"""

    def list_glob_from_store(self):
        """Get list of global config to externally hooked general store"""

    def destroy_glob_from_store(self, name):
        """Destroy global config to externally hooked general store"""

    def commit(self):
        """Write through all saves to store"""

    # Utilities
    def get_object_distribution(self):
        dist = {}
        for i in self.mem.keys():
            t = type(self.mem[i])
            if t in dist.keys():
                dist[t] += 1
            else:
                dist[t] = 1
        return dist
